/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2009 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.docbook.table;

import java.util.Arrays;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.commons.table.operations.ECTableCustomizerDialog;
import ro.sync.ecss.extensions.commons.table.operations.TableInfo;
import ro.sync.ecss.extensions.commons.table.operations.xhtml.SAXHTMLTableCustomizerDialog;

/**
 * Dialog used to customize DocBook table creation. 
 * It is used on Eclipse platform implementation.
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class ECDocbookTableCustomizerDialog extends ECTableCustomizerDialog {
  
  /**
   * Possible values for 'frame' attribute.
   */
  private static final String[] TABLE_FRAME_VALUES = {
    FRAME_NONE,
    FRAME_ALL,
    FRAME_TOP,
    FRAME_BOTTOM,
    FRAME_ABOVE,
    FRAME_BELLOW,
    FRAME_TOPBOT,
    FRAME_SIDES,
    FRAME_HSIDES,
    FRAME_VSIDES,
    FRAME_LHS,
    FRAME_RHS,
    FRAME_BORDER,
    FRAME_BOX,
    FRAME_VOID };
  /**
   * True if a Docbook inner table is inserted
   */
  private final boolean innerTable;

  /**
   * Constructor.
   * 
   * @param parentShell The parent {@link Shell}.
   * @param innerTable <code>true</code> if show for inner table
   */
  public ECDocbookTableCustomizerDialog(org.eclipse.swt.widgets.Shell parentShell, boolean innerTable) {
    super(parentShell, ! innerTable, ! innerTable, !innerTable, false, innerTable);
    this.innerTable = innerTable;
  }
  
  /**
   * Configure Shell. Set a title to it.
   * 
   * @param newShell The new shell.
   * @see org.eclipse.jface.dialogs.Dialog#configureShell(org.eclipse.swt.widgets.Shell)
   */
  @Override
  protected void configureShell(Shell newShell) {
    super.configureShell(newShell);
    if(innerTable) {
      newShell.setText("Insert Entry Table");
    }
  }

  /**
   * @see ro.sync.ecss.extensions.commons.table.operations.ECTableCustomizerDialog#createTitleCheckbox(org.eclipse.swt.widgets.Composite)
   */
  @Override
  protected Button createTitleCheckbox(Composite parent) {
    Button titleCheckBox = new Button(parent, SWT.CHECK);
    titleCheckBox.setText("Title");
    titleCheckBox.setToolTipText("If not checked an 'informaltable' will be inserted.");
    return titleCheckBox;
  }
  
  /**
   * @see ro.sync.ecss.extensions.commons.table.operations.ECTableCustomizerDialog#getColumnWidthsSpecifications(int)
   */
  @Override
  protected List<ColumnWidthsType> getColumnWidthsSpecifications(int tableModel) {
    if (tableModel == TableInfo.TABLE_MODEL_CALS) {
      // CALS table
      return Arrays.asList(CALS_WIDTHS_SPECIFICATIONS);
    } else if (tableModel == TableInfo.TABLE_MODEL_HTML) {
      // HTML table
      return Arrays.asList(HTML_WIDTHS_SPECIFICATIONS);
    }
    return null;
  }
  
  /**
   * Depending on the table model, CALS or HTML, return different values for 'frame' attribute. 
   * 
   * @see ro.sync.ecss.extensions.commons.table.operations.ECTableCustomizerDialog#getFrameValues(int)
   */
  @Override
  protected String[] getFrameValues(int tableModel) {
    String[] frameVals = TABLE_FRAME_VALUES;
    if (tableModel == TableInfo.TABLE_MODEL_HTML) {
      frameVals = SAXHTMLTableCustomizerDialog.TABLE_FRAME_VALUES;
    }
    return frameVals;
  }
  
  /**
   * @see ro.sync.ecss.extensions.commons.table.operations.ECTableCustomizerDialog#getDefaultFrameValue(int)
   */
  @Override
  protected String getDefaultFrameValue(int tableModel) {
    String defaultFrame = FRAME_ALL;
    if (tableModel == TableInfo.TABLE_MODEL_HTML) {
      defaultFrame = FRAME_VOID;
    }
    return defaultFrame;
  }
}