/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2009 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ro.sync.ecss.extensions.api;

import javax.swing.text.BadLocationException;
import javax.swing.text.Position;
import javax.swing.text.Segment;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;

/**
 * Interface to describe a sequence of character content that can be edited.
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PUBLIC)
public interface Content {

	/**
	 * Creates a position within the content. The position offset is changed as the
	 * content is edited.
	 * 
	 * @param offset The offset in the content >= 0
	 * @return A new {@link javax.swing.text.Position}.
	 * @exception BadLocationException For an invalid offset
	 */
	public Position createPosition(int offset) throws BadLocationException;

	/**
	 * The length in characters of the content.
	 * 
	 * @return The length >= 0
	 */
	public int getLength();

  /**
   * Inserts a sequence of characters into the content at a given offset.
   * 
   * @param where
   *            Offset into the content to make the insertion >= 0
   * @param ch The char buffer to insert from. 
   * @param start  Start of useful data in the char buffer.
   * @param length  Length of useful data in the char buffer.
   * @exception BadLocationException
   *                thrown if the offset indicated by the <code>where</code> argument is not
   *                contained in the boundaries of the content character sequence.
   */
  public void insertChars(int where, char[] ch, int start, int length)
      throws BadLocationException;
  
	/**
	 * Removes some portion of the content sequence.
	 * 
	 * @param where
	 *            The offset into the sequence to make the removal >= 0.
	 * @param nitems
	 *            The number of items in the sequence to be removed >= 0.
	 * @exception BadLocationException
	 *                Thrown if the area covered by the <code>where</code> 
	 *                and <code>nitems</code> parameters is not
	 *                contained in the content character sequence.
	 */
	public void remove(int where, int nitems)
			throws BadLocationException;

	/**
	 * Fetches a string of characters contained in the content sequence.
	 * 
	 * @param where
	 *            Offset into the sequence to fetch >= 0.
	 * @param len
	 *            Number of characters to copy >= 0.
	 * @return The extracted String.
	 * @exception BadLocationException
	 *                Thrown if the area covered by the arguments is not
	 *                contained in the content character sequence.
	 */
	public String getString(int where, int len) throws BadLocationException;

  /**
   * Retrieves a portion of the content into the specified {@link Segment}. 
   * If the desired content spans the gap, we copy the content.  
   * If the desired content does not span the gap, the actual store is returned to avoid the copy since
   * it is contiguous.
   *
   * @param where The starting position >= 0, where + len <= length()
   * @param len The number of characters to be retrieved >= 0
   * @param chars The {@link Segment} object to return the characters into.
   * @exception BadLocationException If the specified position or length are invalid.
   */
  public void getChars(int where, int len, Segment chars) throws BadLocationException;
}